/*******************************************************************************
 *
 *        Program : prs.c
 *
 *        Author  : Kim, Dongmin (kimd@top.cis.syr.edu)
 *
 * This program simulates the vector prefix sum and redunction.
 *
 * The vector prefix sum and reduction is like pure prefix sum and reduction
 *
 * except the exchanging data is vector not one variable.
 *
 * In this program, the prefix sum and reduction are combined in one routine
 *
 * so that the communication time is reduced, and
 *
 * two different algorithms are simulated.
 *
 * One is Naive algorithm mentioned in Kuma's Parallel Computing, and
 *
 * the other is Split algorithm.
 *
 * The main difference of these two algorithms is the step and
 *
 * the exchange data size.
 *
 * The Naive has lg N steps and exchanges N size of data.
 *
 * The Split has 2 * lg N steps and exchanges data starting N/2 and
 *
 * reduces the size by half, where N is the vector size.
 *
 * We simulates these two algorithms according to processor number and
 *
 * vector size, and see what happens.
 *
 * Some important procedures are like follows :
 *
 * prs_dim_naive() : Combined prefix and reduction in the Naive algorithm
 *
 * prs_dim_split() : Combined prefix and reduction in the Split algorithm
 *
 *******************************************************************************/

#include <cm/cmmd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "global.h"
#include "function.h"

/*******************************************************************************
 *   Procedure : read_input                                                    *
 *   Input     : None
 *   Output    : None
 *   Function  : read the input variable from the consol.
 *******************************************************************************/
read_input() {
   int error;
   int i, j;
   int total = 1, t_total=1;
   int save;

   ori_nproc = CMMD_partition_size();

   /*---------------- Rank of the input Array ---------------------------*/
   do {
      printf("\nEnter rank of the input array (Max=%d): ", Rankmax);
      scanf("%d", &rank);
      error = ( rank < 1 || rank > Rankmax );
      if (error)
	 printf("\n??? Improper rank. Try again!\n");
    } while (error);

   printf("\nEnter direction to communicate (0 ~ %d): ", rank-1);
   scanf("%d", &direction);

   printf("\nEnter vector size : ");
   scanf("%d", &v_size);

   /*-------------------- from dim=0 to dim=rank-1 ----------------------*/
   for (i=0; i<rank; ++i) {

     /*------------ Source Distribution for A, M ------------------*/
     save = total;
     do {
       printf("\n   Enter number of source processors (Max=%d): ", 
	      Procmax/total);
       scanf("%d", &nproc[i]);
       total *= nproc[i];
       error = ( nproc[i] < 1 || total > ori_nproc );
       dproc[i] = Ilog2(nproc[i]);
       if (error) {
	 total = save;
	 printf("\n??? Improper source processor number. Try again!\n");
       }
     } while (error);
   }

   CMMD_disable_safety();
   CMMD_fset_io_mode(stdin, CMMD_sync_bc);
   CMMD_fset_io_mode(stdout, CMMD_sync_bc);
   CMMD_fset_io_mode(stderr, CMMD_independent);

   tot_nproc = total;
   tot_dproc = Ilog2(tot_nproc);

   /*======= Who I am ? =======*/
   my_pnum  = CMMD_self_address(); 
   map_to_grid(my_pnum, my_grid);

   /* setup the partner, right child and the parent */
   for ( i = 0 ; i < rank ; i++ ) {
     partner_grid[i] = my_grid[i];
     right_grid[i] = my_grid[i];
     parent_grid[i] = my_grid[i];
   }

   i = CMMD_reset_partition_size(tot_nproc);
   if (i != tot_nproc) 
     CMMD_error("\nERROR: CMMD Partition failed!\n");

   /* every processor use a different random seed */
   /*srandom(CMMD_self_address()); */

   if ( CMAML_allocate_this_rport(rport) <= 0 )
     CMMD_error("ERROR: Failed to allocate receive rport %d\n", rport);

}

/*******************************************************************************
 *   Procedure : PRS_dim_naive()
 *   Input     : int dim, int size, int* p_result, int* r_result
 *   Output    : None
 *   Function  : Scalable vector PRS on PS_i & RS_i using naive algorithm        
 *               which is defined Kurmar's book
 *******************************************************************************/
PRS_dim_naive(dim, size, p_result, r_result)
     int dim,          /* PRS along dimension = dim */
         size,         /* vector size = L_(d-1)*L_(d-2)* ... *L_i*T_i */
         *p_result,    /* exclusive prefix */
         *r_result;    /* reduction */
{

  int partner, partner_grid[Rankmax];
  int i, j;

  for (j=0; j<size; ++j) {
    r_result[j] = p_result[j];
    p_result[j] = 0;
  }

  CMAML_disable_interrupts();

  CMAML_set_rport_addr         (rport, in_mesg);
  CMAML_set_rport_handler      (rport, prs_dim_handler);
  CMAML_set_rport_handler_arg2 (rport, null);
  CMAML_set_rport_total_nbytes (rport, msg_size*size);

  CMMD_sync_with_nodes();

  for (j=0; j<rank; ++j)
    partner_grid[j] = my_grid[j];

  for (i=0; i<dproc[dim]; ++i) {
  
    partner_grid[dim] = my_grid[dim] ^ pow2(i);
    map_to_pnum(partner_grid, &partner);

    prs_dim_ack = false;

    CMAML_set_rport_byte_counter (rport, msg_size*size);

    CMMD_sync_with_nodes();

    CMAML_scopy(partner,
		rport,
		0,
		0,
		r_result,
		msg_size*size,
		null,
		null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    for (j=0; j<size; ++j)
      r_result[j] += in_mesg[j];

    if (partner_grid[dim] < my_grid[dim])
      for (j=0; j<size; ++j)
	p_result[j] += in_mesg[j];

    partner_grid[dim] = my_grid[dim];

  }

  CMAML_enable_interrupts();

/****
  show(my_grid[dim], 0, size, p_result);
  show(my_grid[dim], 0, size, r_result);
****/
}

/*******************************************************************************
 *   Procedure : PRS_dim_split()
 *   Input     : int vector_size, int input_v, int* p_result, int* r_result
 *   Output    : None
 *   Function  : Scalable vector PRS on PS_i & RS_i using advanced algorithm
 *******************************************************************************/
PRS_dim_split(vector_size, p_result, r_result)
     int vector_size,  /* vector size = L_(d-1)*L_(d-2)* ... *L_i*T_i */
         *p_result,    /* prefix sum output array */
         *r_result;    /* reduction output array */
{
/*
  int    save[Max];
*/
  int    i, j, k;
  int    step;
  int    size;       /* size of integer */
  int    data_size;  /* data size */
  int    digit;      /* total digit in binary */
  int    check;      /* check if the i-th bit is 1 */
  int    start_send; /* the start index to be sent */
  int    start_recv; /* the start index to be calculated */
  int    start;
  int    save[Max];

/***
  show(my_pnum, 0, vector_size, p_result);
***/

  my_id = my_grid[direction]; 

  dir_nproc  = nproc[direction];

  /* setup the partner, right child and the parent */
  for ( i = 0 ; i < rank ; i++ ) {
    partner_grid[i] = my_grid[i];
    right_grid[i] = my_grid[i];
  }


  /* Only for exclusive */

/****/
  for ( i = 0 ; i < vector_size ; i++ ) {
    save[i] = p_result[i];
  }
/****/
  size = sizeof(int);

  data_size = vector_size;
  start_recv = 0;
  digit = Ilog2(dir_nproc);

  /** For Active Message Passing **/
  CMAML_disable_interrupts();

  CMAML_set_rport_addr         (rport, in_mesg);
  CMAML_set_rport_handler      (rport, prs_dim_handler);
  CMAML_set_rport_handler_arg2 (rport, null);
  /********************************/

  CMMD_sync_with_nodes();
/* ===============
       SUM
   =============== */

  for ( step = 0 ; step < digit ; step++ ) {

    data_size = data_size / 2;
    start_send = start_recv;
    check = my_id & pow2(step);
    if ( check == 0 )
      start_send += data_size;
    else
      start_recv += data_size;

    /* decide the partner and the right child */
    partner_pnum = my_id ^ pow2(step);
    partner_grid[direction] = partner_pnum;
    map_to_pnum(partner_grid, &partner_pnum);

    /** For Active Message Passing **/

    prs_dim_ack = false;

    /* receive */
    CMAML_set_rport_byte_counter (rport, data_size*size);
    CMAML_set_rport_total_nbytes (rport, data_size*size);

    CMMD_sync_with_nodes();

    /* send */
    CMAML_scopy(partner_pnum,
                rport,
                0,
                0,
                &p_result[start_send],
                data_size*size,
                null,
                null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    /********************************/

/***
    for ( i = start_recv ; i < start_recv + data_size ; i++ ) {
      Right[step][i] = p_result[i];
      p_result[i] += in_mesg[i];
    }
***/

    if ( my_id < partner_pnum ) {
      for ( i = 0 ; i < data_size ; i++ ) {
	p_result[i + start_recv] += in_mesg[i];
	Right[step][i + start_recv] = in_mesg[i];
      }
    }
    else {
      for ( i = 0 ; i < data_size ; i++ ) {
	Right[step][i + start_recv] = p_result[i + start_recv];
	p_result[i + start_recv] += in_mesg[i];
      }
    }
  }

  /* copy the p_result to r_result for the reduction */
  for ( i = 0 ; i < vector_size ; i++ )
    r_result[i] = p_result[i];

/****
  show(my_pnum, 0, vector_size, p_result);
  show(my_pnum, 0, vector_size, Right[0]);
****/

  start_send = start_recv;

/* =======================
     PRESFIX & REDUCTION
   ======================= */

  for ( step = digit - 1 ; step >= 0 ; step-- ) {

    check = my_id & pow2(step);
    if ( check == 0 )
      start_recv += data_size;
    else
      start_recv -= data_size;

    /* decide the partner */
    partner_pnum = my_id ^ pow2(step);
    partner_grid[direction] = partner_pnum;
    map_to_pnum(partner_grid, &partner_pnum);

    if ( my_id < partner_pnum ) {
      for ( i = 0 ; i < data_size ; i++ ) {
	out_mesg[i] = p_result[i + start_send];
	p_result[i + start_send] -= Right[step][i + start_send];
	out_mesg[i + data_size] = r_result[i + start_send];
      }
    }
    else {
      for ( i = 0 ; i < data_size ; i++ ) {
	out_mesg[i] = p_result[i + start_send] - Right[step][i + start_send];
	out_mesg[i + data_size] = r_result[i + start_send];
      }
    }

    /** For Active Message Passing **/

    prs_dim_ack = false;

    /* receive */
/***
    CMAML_set_rport_addr         (rport, in_mesg);
***/
    CMAML_set_rport_byte_counter (rport, data_size*size*2);
    CMAML_set_rport_total_nbytes (rport, data_size*size*2);

    CMMD_sync_with_nodes();

    /* send */
    CMAML_scopy(partner_pnum,
		rport,
		0,
		0,
		out_mesg,
		data_size*size*2,
		null,
		null);  

    CMAML_poll_while_lt(&prs_dim_ack, true); 

    /********************************/

    for ( i = 0 ; i < data_size ; i++ ) {
      p_result[i + start_recv] = in_mesg[i];
      r_result[i + start_recv] = in_mesg[i + data_size];
    }

    data_size = data_size * 2;

    if ( start_send > start_recv )
      start_send = start_recv;
    start_recv = start_send;
  }

/* For exclusive prefix */
/****/
  for ( i = 0 ; i < vector_size ; i++ )
    p_result[i] -= save[i];
/****/
/****
  show(my_pnum, 0, vector_size, p_result);
  show(my_pnum, 0, vector_size, r_result);
****/

}

/*******************************************************************************
 *                                                                             * 
 *                              MAIN  PROGRAM                                  *
 *                                                                             *
 *******************************************************************************/
  
main(argc, argv)
     int argc;
     char *argv[];
{
  FILE   *fd; /* input vector file name */

  int    i, j, k;
  int    step;
  int    size;       /* size of integer */
  int    data_size;  /* data size */
  int    digit;      /* total digit in binary */
  int    check;      /* check if the i-th bit is 1 */
  int    start_send; /* the start index to be sent */
  int    start_recv; /* the start index to be calculated */
  int    start;
  double time, time0, time1, time2;
  int    p_result[Max], r_result[Max];
  int    input_v[Max];
  double cmmd, prs;

  /*-------------------------  Initialize  -----------------------------*/

  read_input();

  my_id = my_grid[direction]; 

  dir_nproc  = nproc[direction];

/* Initialize the vector in each processor */

  for ( i = 0 ; i < v_size ; i++ ) {
    p_result[i]  = 2;
  }

  CMMD_node_timer_clear(0);
  CMMD_node_timer_clear(1);

  CMMD_node_timer_start(1);
  PRS_dim_naive(direction, v_size, p_result, r_result);
  CMMD_node_timer_stop(1);

  for ( i = 0 ; i < v_size ; i++ ) {
    p_result[i]  = 2;
  }

  CMMD_node_timer_start(0);
  PRS_dim_split(v_size, p_result, r_result);
  CMMD_node_timer_stop(0);

  time0 = CMMD_node_timer_busy(0) * 1000.;  /* milisec */
  time1 = CMMD_node_timer_busy(1) * 1000.;  /* milisec */

  CMMD_fset_io_mode(stdout, CMMD_sync_bc);
  printf("\n\n*********************************************************\n");
  printf("\n  Nproc       = %d", tot_nproc);
  printf("\n  Vector size = %d\n", v_size);
  printf("\n  CMMD Naive Algorithm     = %5.3f msec", time1);
  printf("\n  CMMD Split Algorithm     = %5.3f msec", time0);
  printf("\n\n*********************************************************\n\n");

}
